@@ -1,7 +1,7 @@
 from collections import OrderedDict
 
 from _pytest.runner import CollectReport
-from py.log import Producer
+from xdist.remote import Producer
 from xdist.report import report_collection_diff
 from xdist.workermanage import parse_spec_config
 
@@ -361,12 +361,12 @@ class LoadScopeScheduling:
         extra_nodes = len(self.nodes) - len(self.workqueue)
 
         if extra_nodes > 0:
-            self.log("Shuting down {} nodes".format(extra_nodes))
+            self.log("Shutting down {} nodes".format(extra_nodes))
 
             for _ in range(extra_nodes):
                 unused_node, assigned = self.assigned_work.popitem(last=True)
 
-                self.log("Shuting down unused node {}".format(unused_node))
+                self.log("Shutting down unused node {}".format(unused_node))
                 unused_node.shutdown()
 
         # Assign initial workload
